/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#define VERSION_REV 17
#define VERSION_DATE "2017-07-03"

#include "flxstat.h"
#include "flxfunction_ope_calc.h"
#include <fstream>

#if defined(_MSC_VER)
  #include "flxlibstat_export.h"
#else
  #define FLXLIBSTAT_EXPORT
#endif

#ifdef __WINDOWS__
  #pragma comment(lib, "flxlib")
#endif

flxlibs_interface flx_iface;
FlxData* flxdata_ref = NULL;
flxStatBox_container StatBox_cnt;
flx_cdf_container CDFbox;

#if FLX_DEBUG
int flxStatBox_container::Cinst = 0;
#endif


extern "C" FLXLIBSTAT_EXPORT void flxlib_init( FlxObjectReadBox* objReadBox, FlxData* dataBox ) {
  GlobalVar.slog(4) << "  Stat-library - Version " << VERSION_REV << " (" << VERSION_DATE << ")" << std::endl;
  flx_iface.set_boxes(objReadBox,&(dataBox->FunBox));
  // objects
    flx_iface.add_obj("stat_box_new", new FlxObjReadStat_Box_new()); 
    flx_iface.add_obj("stat_box_add", new FlxObjReadStat_Box_add()); 
    flx_iface.add_obj("stat_box_vec_mean", new FlxObjReadStat_Box_vec_mean()); 
    flx_iface.add_obj("stat_box_confi", new FlxObjReadStat_Box_confi());  
    flx_iface.add_obj("stat_cdf_new", new FlxObjReadStat_CDF_new()); 
    flx_iface.add_obj("stat_cdf_append", new FlxObjReadStat_CDF_append()); 
    flx_iface.add_obj("stat_cdf_plot", new FlxObjReadStat_CDF_plot()); 
    flx_iface.add_obj("stat_cdf_save", new FlxObjReadStat_CDF_save()); 
    flx_iface.add_obj("stat_cdf_load", new FlxObjReadStat_CDF_load()); 
  // functions
    flxdata_ref = dataBox;
}

extern "C" FLXLIBSTAT_EXPORT void flxlib_close() {
  flx_iface.free_memory();
}

// ------------------------------------------------------------------------------------------------

flxStatBox_container::flxStatBox_container()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'flxStatBox_container' created ...";
      throw FlxException("flxStatBox_container::flxStatBox_container", ssV.str() );
    }
  #endif
}

flxStatBox_container::~flxStatBox_container()
{
  for (std::map<std::string, FlxStatBox*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    delete pos->second;
  }
}

void flxStatBox_container::insert(const std::string name, const tuint N, const tuint M)
{
  std::map<std::string, FlxStatBox*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    FlxStatBox* sb = pos->second;
    if (sb->N==N && sb->M==M) {
      sb->clear();
    } else {
      delete sb;
      pos->second = new FlxStatBox(N,M);
    }
  } else {
    FlxStatBox* valuep = new FlxStatBox(N,M);
    std::pair<std::string, FlxStatBox*> Element(name, valuep);
    box.insert(Element);
  }
}

FlxStatBox& flxStatBox_container::get(const std::string& name)
{
  std::map<std::string, FlxStatBox*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return (*pos->second);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "The stat-box '" << name << "' does not exist.";
    throw FlxException("flxStatBox_container::get_1", ssV_2.str(), "You have to define the container before you can use it."); 
  }
}


// ------------------------------------------------------------------------------------------------

FlxObjStat_Box_new::~FlxObjStat_Box_new()
{
  delete fname;
  delete M;
  delete N;
}

void FlxObjStat_Box_new::task()
{
  const tuint Ni = N->cast2tuint(false);
  const tuint Mi = M->cast2tuint(false);
  const std::string name = fname->eval_word(true);
  StatBox_cnt.insert(name,Ni,Mi);
  GlobalVar.slog(4) << "stat: defined new container '" << name << "' with N=" << Ni << " and M=" << Mi << std::endl;
}

FlxObjBase* FlxObjReadStat_Box_new::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxFunction* N = NULL;
  FlxFunction* M = NULL;
  try {
    reader->getChar('(');
    N = new FlxFunction(funReader,false);
    reader->getChar(',');
    M = new FlxFunction(funReader,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_Box_new(get_doLog(),fname,N,M);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_new::read",1);
    delete fname;
    if (N) delete N;
    if (M) delete M;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

FlxObjStat_Box_add::~FlxObjStat_Box_add()
{
  delete fname;
  delete vecName;
}

void FlxObjStat_Box_add::task()
{
  const std::string name = fname->eval_word(true);
  FlxStatBox& sb = StatBox_cnt.get(name);
  tuint N = 0;
  tdouble* tvp = data->ConstMtxBox.get_Vec(vecName->eval_word(true),N);
  flxVec tv(tvp,N);
  sb.add(tv);
}

FlxObjBase* FlxObjReadStat_Box_add::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* vecName = NULL;
  try {
    reader->getChar('(');
    vecName = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_Box_add(get_doLog(),fname,vecName);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_add::read",1);
    delete fname;
    if (vecName) delete vecName;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

void FlxObjStat_Box_vec_mean::task()
{
  const std::string name = fname->eval_word(true);
  FlxStatBox& sb = StatBox_cnt.get(name);
  tuint Nc = sb.M;
  tdouble* tvp = data->ConstMtxBox.get_Vec(vecName->eval_word(true),Nc);
  flxVec tv(tvp,Nc);
  sb.compute_mean(tv);
}

FlxObjStat_Box_vec_mean::~FlxObjStat_Box_vec_mean()
{
  delete fname;
  delete vecName;
}

FlxObjBase* FlxObjReadStat_Box_vec_mean::read()
{
  FlxString* vecName = new FlxString(false,false);
  FlxString* fname = NULL;
  try {
    reader->getChar('=');
    fname = new FlxString(false,false);
    read_optionalPara(false);
    return new FlxObjStat_Box_vec_mean(get_doLog(),fname,vecName);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_vec_mean::read",1);
    delete vecName;
    if (fname) delete fname;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

void FlxObjStat_Box_confi::task()
{
  const std::string name = fname->eval_word(true);
  FlxStatBox& sb = StatBox_cnt.get(name);
  tuint N = 0;
  flxVec clevels(data->ConstMtxBox.get_Vec(clf->eval(),N),N);
  if (userColName.empty()) {
    sb.plot_confidence(sout(),clevels,fixW,NULL);
  } else {
    N = 0;
    tdouble* userCol = data->ConstMtxBox.get_Vec(userColName,N);
    flxVec uc(userCol,N);
    sb.plot_confidence(sout(),clevels,fixW,&uc);
  }
}

FlxObjReadStat_Box_confi::FlxObjReadStat_Box_confi()
 : FlxObjReadOutputBase(true) 
{
  AllDefParaBox->insert(new FlxOptionalParaFlxString("","stat_box_confi::usercol",false));
  ParaBox.insert("usercol", "stat_box_confi::usercol" );
}

FlxObjBase* FlxObjReadStat_Box_confi::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxMtxConstFun* clf = NULL;
  try {
    reader->getChar('(');
    clf = new FlxMtxConstFun(true);
    reader->getChar(')');
    read_optionalPara(false);
    FlxString* uc = get_optPara_FlxString("usercol");
    std::string uce("");
    try {
      uce = uc->eval_word(true,true);
      delete uc; uc = NULL;
    } catch (FlxException& e) {
      FLXMSG("FlxObjReadStat_Box_confi::read_1",1);
      delete uc;
      throw;
    }
    return new FlxObjStat_Box_confi(get_doLog(),get_stream(),fname,clf,get_fixW(),uce);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_confi::read_2",1);
    delete fname;
    if (clf) delete clf;
    throw;
  }  
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------

void flx_cdf_container::insert(const std::string& name, const size_t Nreserve)
{
  std::map<std::string, flx_cdf*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    throw FlxException("flx_cdf_container::insert", "A data-CDF with name '" + name + "' is already declared.");
  } else {
    flx_cdf* valuep = new flx_cdf(Nreserve);
    std::pair<std::string, flx_cdf*> Element(name, valuep);
    box.insert(Element);
  }
}

flx_cdf& flx_cdf_container::get(const std::string& name)
{
  std::map<std::string, flx_cdf*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return (*pos->second);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "The data-CDF '" << name << "' does not exist.";
    throw FlxException("flx_cdf_container::get_1", ssV_2.str(), "You have to define the data-CDF before you can use it."); 
  }
}

flx_cdf::flx_cdf(const size_t Nreserve)
: N(0)
{
  list.reserve(Nreserve);
}

void flx_cdf::append(flxVec& dV)
{
  size_t i = 0;		// counter variable
  for (size_t j = 0; j<dV.get_N(); ++j) {	// loop over each number to append to the data-CDF
    const float dVe = dV[j];
    while ( (i<list.size()) && (list[i].value<dVe) ) {		// jump to position in list that has relevant value
      ++i;
    };
    ++N;
    if ( (i<list.size()) && (list[i].value == dVe) ) {	// increase counter of list[i]
      ++(list[i].occ);
    } else {				// insert new entry before list[i]
      flx_cdf_entry e = { float(dVe), 1 };
      list.insert(list.begin()+i,e);
    }
  };
}

void flx_cdf::plot(std::ostream& sout, const bool checkTOL, const int prec, int fixW)
{
  tdouble pdxl = -log(tdouble(N));
  size_t count = 0;
  tdouble px = ZERO;
  tdouble qpx = ONE;
  size_t Nlist = list.size();
  for (size_t i=0;i<Nlist;++i) {
    sout << GlobalVar.Double2String(list[i].value,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(px,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(qpx,checkTOL,prec,fixW) << std::endl;
    count += list[i].occ;
    px = exp(pdxl + log(tdouble(count)));
    qpx = exp(pdxl + log(tdouble(N-count)));
    if (i+1==Nlist) qpx = ZERO;
    sout << GlobalVar.Double2String(list[i].value,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(px,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(qpx,checkTOL,prec,fixW) << std::endl;
  };
}

void flx_cdf::save(const std::string& fileS)
{
  std::ofstream file(fileS.c_str(), std::ios::out | std::ios::binary);
  if (file.is_open()==false) {
    throw FlxException("flx_cdf::save", "File '" + fileS + "' could not be opened.");
  }
  file.write((char*)&(list[0]),sizeof(flx_cdf_entry)*list.size());
  file.close();
}

void flx_cdf::load(const std::string& fileS)
{
  if (N>0) throw FlxException("flx_cdf::load_01", "A stored data-CDF can only be loaded to an empty data-CDF.");
  std::ifstream file(fileS.c_str(), std::ios::in | std::ios::binary | std::ios::ate );
  if (file.is_open()==false || file.good()==false) {
    throw FlxException("flx_cdf::load_02", "File '" + fileS + "' could not be opened.");
  }
  std::streampos size = file.tellg();
  size_t sizeN = size/sizeof(flx_cdf_entry);
  if ( (size_t)size != sizeN*sizeof(flx_cdf_entry) ) {
    throw FlxException("flx_cdf::load_03", "The file '" + fileS + "' seems to be either corrupt or does not contain a data-CDF.");
  }
  list.resize(sizeN);
  file.seekg (0, std::ios::beg);
  file.read((char*)&(list[0]),size);
  file.close();
  // count N
    size_t Nlist = list.size();
    for (size_t i=0;i<Nlist;++i) {
      N += list[i].occ;
    };
}

void FlxObjStat_CDF_new::task()
{
  const size_t Nreserve = Nreservef->cast2tulong(false);
  const std::string name = fname->eval_word(true);
  CDFbox.insert(name,Nreserve);
  GlobalVar.slog(4) << "stat: defined new data-CDF '" << name << "' with capacity for " << Nreserve << " entries." << std::endl;
}

FlxObjReadStat_CDF_new::FlxObjReadStat_CDF_new()
: FlxObjReadBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e8,"stat_cdf_new::nreserve"));
  ParaBox.insert("nreserve", "stat_cdf_new::nreserve" );
}

FlxObjBase* FlxObjReadStat_CDF_new::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxFunction* Nreserve = NULL;
  try {
    read_optionalPara(false);
    Nreserve = get_optPara_FlxFunction("nreserve");
    return new FlxObjStat_CDF_new(get_doLog(),fname,Nreserve);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_new::read",1);
    delete fname;
    if (Nreserve) delete Nreserve;
    throw;
  }
}

void FlxObjStat_CDF_append::task()
{
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  // get input stream
    const std::string isn = istrmf->eval_word(true);
    FlxIstream &is = data->IstreamBox.get(isn);
  // bulk-size
    size_t bN = bulkf->cast2tuint(false);
  flxVec dV(bN);
  tdoublePtr dP = dV.get_tmp_vptr();
  tuint lastIndex = 0;
  bool b;
  do {
    b = is.get_vec(dV,lastIndex,true);
    // transform the vector
      tdouble &x_ref = data->ConstantBox.getRef("x");
      const tdouble x_old = x_ref;
      for (size_t i=0;i<lastIndex;++i) {
	x_ref = dP[i];
	dP[i] = transform->calc();
      };
      x_ref = x_old;
    std::sort(dP,dP+lastIndex);
    flxVec tV(dP,lastIndex);
    dcdf.append(tV);
  } while (b);
}

FlxObjReadStat_CDF_append::FlxObjReadStat_CDF_append()
: FlxObjReadBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e7,"stat_cdf_append::bulk"));
  ParaBox.insert("bulk", "stat_cdf_append::bulk" );
  AllDefParaBox->insert(new FlxOptionalParaFun(new FlxFunction(new FunConst(data->ConstantBox.get("x",true))),"stat_cdf_append::transform"));
  ParaBox.insert("transform", "stat_cdf_append::transform" );
}

FlxObjBase* FlxObjReadStat_CDF_append::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* istrmf = NULL;
  try {
    reader->getChar('(',false);
    istrmf = new FlxString(false,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjStat_CDF_append(get_doLog(), fname, istrmf, get_optPara_FlxFunction("bulk"), get_optPara_FlxFunction("transform"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadStat_CDF_append::read",1);
    delete fname;
    if (istrmf) delete istrmf;
    throw;
  }
}

void FlxObjStat_CDF_plot::task()
{
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  dcdf.plot(sout(),checkTOL,prec,fixW);
}

FlxObjBase* FlxObjReadStat_CDF_plot::read()
{
  FlxString* fname = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjStat_CDF_plot(get_doLog(),get_stream(),fname,get_prec(),get_fixW());
  } catch (FlxException &e) {
    delete fname;
    throw;
  }
}

void FlxObjStat_CDF_save::task()
{
  const std::string name = fname->eval_word(true);
  const std::string fileS = fileF->eval(false);
  flx_cdf& dcdf = CDFbox.get(name);
  dcdf.save(fileS);
  GlobalVar.slog(4) << "stat: stored data-CDF '" << name << "' in file '" << fileS << "'." << std::endl;
}

FlxObjBase* FlxObjReadStat_CDF_save::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* fileF = NULL;
  try {
    reader->getChar('(');
    fileF = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_CDF_save(get_doLog(),fname,fileF);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_save::read",1);
    delete fname;
    if (fileF) delete fileF;
    throw;
  }
}

void FlxObjStat_CDF_load::task()
{
  const std::string name = fname->eval_word(true);
  const std::string fileS = fileF->eval(false);
  flx_cdf& dcdf = CDFbox.get(name);
  dcdf.load(fileS);
  GlobalVar.slog(4) << "stat: loaded data-CDF '" << name << "' from file '" << fileS << "'." << std::endl;
}

FlxObjBase* FlxObjReadStat_CDF_load::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* fileF = NULL;
  try {
    reader->getChar('(');
    fileF = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_CDF_load(get_doLog(),fname,fileF);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_load::read",1);
    delete fname;
    if (fileF) delete fileF;
    throw;
  }
}



